"""
Module with classes for tag selection based on CSS selectors.

Contains implementation of basic CSS pseudo-classes like
:only-child, :empty, :nth-child().

They can be used in combination with other SoupSelector objects
to create more complex tag selection conditions.

This module contains the following classes:
* OnlyChild
* Empty
* FirstChild
* LastChild
* NthChild
* NthLastChild
* FirstOfType
* LastOfType
* NthOfType
* NthLastOfType
* OnlyOfType
* CSS - wrapper for simple search with CSS selectors
"""

from itertools import islice
from typing import Optional

import soupsieve as sv
from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.selectors.base import SelectableCSS, SoupSelector
from soupsavvy.selectors.tag_utils import TagIterator


class CSSSoupSelector(SoupSelector, SelectableCSS):
    """
    Base class for CSS selector based tag selection.
    Classes that base their selection on CSS selectors should inherit from this class
    and implement the `selector` property.

    By default, the `find` methods are implemented using the soupsieve library
    to match the selector.

    CSSSoupSelector objects are based on SoupSelector interface
    and can be easily used in combination with other SoupSelector objects.
    """

    _SELECTOR: str

    def __init__(self) -> None:
        selector = self.__class__._SELECTOR.format(*self._formats)

        try:
            self._compiled = sv.compile(selector)
        except sv.SelectorSyntaxError:
            raise exc.InvalidCSSSelector(
                "CSS selector constructed from provided parameters "
                f"is not valid: {selector}"
            )
        self._selector = selector

    @property
    def _formats(self) -> list[str]:
        """
        List of arguments to be used to format css selector string of the selector.
        """
        return []

    @property
    def css(self) -> str:
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:

        iterator = TagIterator(tag, recursive=recursive)
        return list(islice(filter(self._compiled.match, iterator), limit))

    def __eq__(self, other: object) -> bool:
        # we only care if selector is equal with current implementation
        return isinstance(other, CSSSoupSelector) and self.css == other.css


class OnlyChild(CSSSoupSelector):
    """
    Class to select tags that are the only child of their parent.
    It uses the CSS selector `:only-child`.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ❌
    ...    <div class="menu">Hello World 2 </div> ❌
    ... </div>

    Tag is only selected if it is the only child of its parent.

    Example
    --------
    >>> OnlyChild().selector
    :only-child

    For more information on the :only-child selector, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-child
    """

    _SELECTOR = ":only-child"


class Empty(CSSSoupSelector):
    """
    Class to select tags that are empty, i.e., have no children.
    It uses the CSS selector `:empty`.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu"></div> ✔️
    ... </div>

    Tag is only selected if it is empty.

    Example
    --------
    >>> Empty().selector
    :empty

    Notes
    --------
    Any text or whitespace inside the tag is consider as a child
    and makes the tag non-empty.

    Example
    --------
    >>> <div class="widget">Hello World</div> ❌
    ... <div class="widget"> </div> ❌

    These tags are not empty and do not match the selector.

    For more information on the :empty selector, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:empty
    """

    _SELECTOR = ":empty"


class FirstChild(CSSSoupSelector):
    """
    Class to select tags that are the first child of their parent.
    It uses the CSS selector `:first-child`.

    Example
    --------
    >>> <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <div class="menu">Hello World 2 </div> ❌
    ... </div>

    Tag is only selected if it is the first child of its parent.

    Example
    --------
    >>> FirstChild().selector
    :first-child

    Notes
    --------
    FirstChild object is essentially the same as NthChild("1").

    For more information on the :first-child selector, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:first-child
    """

    _SELECTOR = ":first-child"


class LastChild(CSSSoupSelector):
    """
    Class to select tags that are the last child of their parent.
    It uses the CSS selector `:last-child`.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ❌
    ...    <div class="menu">Hello World 2 </div> ✔️
    ... </div>

    Tag is only selected if it is the last child of its parent.
    Element that is the first and only child is matched as well.

    Example
    --------
    >>> LastChild().selector
    :last-child

    Notes
    --------
    LastChild object is essentially the same as NthLastChild("1").

    For more information on the :last-child selector, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:last-child
    """

    _SELECTOR = ":last-child"


class _NthBaseSelector(CSSSoupSelector):
    """Base class for selectors based on nth formula."""

    def __init__(self, nth: str) -> None:
        self._nth = nth
        super().__init__()

    @property
    def _formats(self) -> list[str]:
        return [self._nth]


class NthChild(_NthBaseSelector):
    """
    Class to select tags that are the nth child of their parent.
    It uses the CSS selector `:nth-child(n)`.

    Parameters
    ----------
    nth : str, positional
        Number of the child to be selected. Can be a number or a formula.

    Example
    --------
    >>> NthChild("2").selector
    :nth-child(2)
    >>> NthChild("2n+1").selector
    :nth-child(2n+1)
    >>> NthChild("odd").selector
    :nth-child(odd)

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-child
    """

    _SELECTOR = ":nth-child({})"


class NthLastChild(_NthBaseSelector):
    """
    Class to select tags that are the nth last child of their parent.
    It uses the CSS selector `:nth-last-child(n)`.

    Parameters
    ----------
    nth : str, positional
        Number of the child to be selected. Can be a number or a formula.

    Example
    --------
    >>> NthLastChild("2").selector
    :nth-last-child(2)
    >>> NthLastChild("2n+1").selector
    :nth-last-child(2n+1)
    >>> NthLastChild("odd").selector
    :nth-last-child(odd)

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-child
    """

    _SELECTOR = ":nth-last-child({})"


class FirstOfType(CSSSoupSelector):
    """
    Class to select tags that are the first of their type in their parent.
    It uses the CSS selector `:first-of-type`.

    Example
    --------
    >>> <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <span class="menu">Hello World 2 </span> ✔️
    ...    <div class="menu">Hello World 3 </div> ❌
    ... </div>

    Tag is only selected if it is the first of its type in its parent.

    Example
    --------
    >>> FirstOfType().selector
    :first-of-type

    Notes
    --------
    For this selector the first tag of any type is selected, which in
    case of finding single tag is equivalent to FirstChild() results.

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:first-of-type
    """

    _SELECTOR = ":first-of-type"


class LastOfType(CSSSoupSelector):
    """
    Class to select tags that are the last of their type in their parent.
    It uses the CSS selector `:last-of-type`.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ❌
    ...    <span class="menu">Hello World 2 </span> ✔️
    ...    <div class="menu">Hello World 3 </div> ✔️
    ... </div>

    Tag is only selected if it is the last of its type in its parent.
    Element that is the first and only child is matched as well.

    Example
    --------
    >>> LastOfType().selector
    :last-of-type

    Notes
    --------
    For this selector the last tag of any type is selected, which in
    case of finding single tag is the equivalent to LastChild() results.

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:last-of-type
    """

    _SELECTOR = ":last-of-type"


class NthOfType(_NthBaseSelector):
    """
    Class to select tags that are the nth of their type in their parent.
    It uses the CSS selector `:nth-of-type(n)`.

    Parameters
    ----------
    nth : str, positional
        Number of the tag to be selected. Can be a number or a formula.

    Example
    --------
    >>> NthOfType("2").selector
    :nth-of-type(2)
    >>> NthOfType("2n+1").selector
    :nth-of-type(2n+1)
    >>> NthOfType("even").selector
    :nth-of-type(even)

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-of-type
    """

    _SELECTOR = ":nth-of-type({})"


class NthLastOfType(_NthBaseSelector):
    """
    Class to select tags that are the nth last of their type in their parent.
    It uses the CSS selector `:nth-last-of-type(n)`.

    Parameters
    ----------
    nth : str, positional
        Number of the tag to be selected. Can be a number or a formula.

    Example
    --------
    >>> NthLastOfType("2").selector
    :nth-last-of-type(2)
    >>> NthLastOfType("2n+1").selector
    :nth-last-of-type(2n+1)
    >>> NthLastOfType("even").selector
    :nth-last-of-type(even)

    For more information on the formula, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-of-type
    """

    _SELECTOR = ":nth-last-of-type({})"


class OnlyOfType(CSSSoupSelector):
    """
    Class to select tags that are the only of their type in their parent.
    It uses the CSS selector `:only-of-type`.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <span class="menu">Hello World 2 </span> ❌
    ...    <span class="menu">Hello World 3 </span> ❌
    ... </div>
    ... <a class="widget"></a> ✔️

    Tag is only selected if it is the only tag of its type in its parent.

    Example
    --------
    >>> OnlyOfType().selector
    :only-of-type

    For more information on the :only-of-type selector, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-of-type
    """

    _SELECTOR = ":only-of-type"


class CSS(CSSSoupSelector):
    """
    Soupsavvy wrapper for simple search with CSS selectors.
    Uses soupsieve library to match the tag, based on the provided CSS selector.

    Extends bs4 Tag.select implementation by adding non recursive search option.
    Provided css selector might be any selector supported by soupsieve.

    Parameters
    ----------
    css : str
        CSS selector to be used for tag selection.

    Example
    --------
    >>> CSS("div.menu")

    Would match:

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="menu_main"> ❌
    ...    <a class="menu">Hello World</a> ❌
    ... </div>
    ... <div class="menu"></div> ✔️
    """

    _SELECTOR = "{}"

    def __init__(self, css: str) -> None:
        self._css = css
        super().__init__()

    @property
    def _formats(self) -> list[str]:
        return [self._css]



"""Helper module with constants for the HTML generator module."""

VOID_TAGS = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "source",
    "track",
    "wbr",
]



"""Module with configuration for the templates."""

from soupsavvy.testing.generators.templates.base import BaseTemplate
from soupsavvy.testing.generators.templates.templates import EmptyTemplate

# Default template used in AttributeGenerator when no value is specified
DEFAULT_VALUE_TEMPLATE: BaseTemplate = EmptyTemplate()

# Default template used in TagGenerator when no text is specified
DEFAULT_TEXT_TEMPLATE: BaseTemplate = EmptyTemplate()



"""
Module with nth-of-type selector implementations for SoupSelectors.
Searches for nth-of-type element in similar fashion to CSS nth-of-type selector,
but 'type' is any SoupSelector instance.
"""

from typing import Optional

from bs4 import Tag

from soupsavvy.selectors.base import SoupSelector, check_selector
from soupsavvy.selectors.nth.nth_utils import parse_nth
from soupsavvy.selectors.tag_utils import TagIterator, TagResultSet


class _BaseNthOfSelector(SoupSelector):
    """
    Base class for nth-of-selector and nth-last-of-selector
    that implements general logic for finding matching elements.
    """

    # slice for modification of list of elements matching selector
    _slice: slice

    def __init__(self, selector: SoupSelector, nth: str) -> None:
        """
        Initializes NthOfSelector instance

        Parameters
        ----------
        selector : SoupSelector
            Any SoupSelector instance to match tags.
        nth : str
            CSS nth selector string. Accepts all valid css nth selectors.

        Raises
        ------
        NotSoupSelectorException
            If selector is not an instance of SoupSelector.
        """
        self._selector = check_selector(selector)
        self.nth_selector = parse_nth(nth)

    @property
    def selector(self) -> SoupSelector:
        """
        Returns selector instance used for matching elements in this nth selector.

        Returns
        -------
        SoupSelector
            Selector used in this nth selector.
        """
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        # if recursive is False, check only children of tag itself
        tag_iterator = (
            TagIterator(tag, recursive=recursive, include_self=True)
            if recursive
            else iter([tag])
        )

        matches = []

        for tag_ in tag_iterator:
            matching = self.selector.find_all(tag=tag_, recursive=False)[self._slice]
            matches += [
                matching[index - 1]
                for index in self.nth_selector.generate(len(matching))
            ]

        # keep order of tags and limit
        results = TagResultSet(
            list(TagIterator(tag, recursive=recursive))
        ) & TagResultSet(matches)

        return results.fetch(limit)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (
            self.selector == other.selector and self.nth_selector == other.nth_selector
        )

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}(selector={self.selector}, nth={self.nth_selector})"

    def __str__(self):
        return repr(self)


class NthOfSelector(_BaseNthOfSelector):
    """
    Component that matches nth-of-type elements in the soup, with use of standard
    css nth selectors. Element is of type if it matches provided SoupSelector instance.

    Example
    -------
    >>> selector = NthOfSelector(AttributeSelector("class", "item"), "2n+1")

    matches all odd elements with class "item".

    Example
    -------
    >>> <div class="item">1</div> ✔️
    >>> <div id="item"></div> ❌
    >>> <div class="item">2</div> ❌
    >>> <div class="item">3</div> ✔️
    >>> <div class="widget"></div> ❌
    >>> <div class="item">4</div> ❌

    For more information about standard :nth-of-type selector, visit:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-of-type

    and check out soupsavvy implementation of :nth-of-type selector:
    `soupsavvy.tags.css.tag_selectors.NthOfType`
    """

    # keep initial order of matching elements
    _slice = slice(None)


class NthLastOfSelector(_BaseNthOfSelector):
    """
    Component that matches nth-last-of-type elements in the soup, with use of standard
    css nth selectors. Element is of type if it matches provided SoupSelector instance.

    Example
    -------
    >>> selector = NthLastOfSelector(AttributeSelector("class", "item"), "2n+1")

    matches all odd elements with class "item" starting from the last element.

    Example
    -------
    >>> <div class="item">1</div> ❌
    >>> <div id="item"></div> ❌
    >>> <div class="item">2</div> ✔️
    >>> <div class="item">3</div> ❌
    >>> <div class="widget"></div> ❌
    >>> <div class="item">4</div> ✔️

    For more information about standard :nth-of-type selector, visit:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-of-type

    and check out soupsavvy implementation of :nth-of-type selector:
    `soupsavvy.tags.css.tag_selectors.NthLastOfType`
    """

    # reverse order of matching elements
    _slice = slice(None, None, -1)


class OnlyOfSelector(SoupSelector):
    """
    Component that matches only-of-type elements in the soup. Element is of type
    if it matches provided SoupSelector instance.

    Example
    -------
    >>> selector = NthLastOnlyOfSelectorOfSelector(AttributeSelector("class", "item"))

    matches all elements with class "item" that are the only child of their parent
    that matches the selector.

    Example
    -------
    >>> <div><div class="item"></div><a class="item"></a></div> ❌
    >>> <div><div class="item"></div><a class="widget"></a></div> ✔️
    >>> <div><div class="item"></div></div> ✔️
    >>> <div><div class="widget"></div></div> ❌

    For all positive examples above, `<div class="item">` is matched.

    For more information about standard :only-of-type selector, visit:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-of-type

    and check out soupsavvy implementation of :only-of-type selector:
    `soupsavvy.tags.css.tag_selectors.OnlyOfType`
    """

    def __init__(self, selector: SoupSelector) -> None:
        """
        Initializes OnlyOfSelector instance.

        Parameters
        ----------
        selector : SoupSelector
            Any SoupSelector instance to match tags.

        Raises
        ------
        NotSoupSelectorException
            If selector is not an instance of SoupSelector.
        """
        self.selector = check_selector(selector)

    def find_all(self, tag: Tag, recursive: bool = True, limit=None) -> list[Tag]:
        tag_iterator = (
            TagIterator(tag, recursive=recursive, include_self=True)
            if recursive
            else iter([tag])
        )

        matching = [
            self.selector.find_all(tag=tag_, recursive=False) for tag_ in tag_iterator
        ]
        matches = [elements[0] for elements in matching if len(elements) == 1]

        # keep order of tags and limit
        results = TagResultSet(
            list(TagIterator(tag, recursive=recursive))
        ) & TagResultSet(matches)

        return results.fetch(limit)

    def __eq__(self, other):
        if not isinstance(other, OnlyOfSelector):
            return False

        return self.selector == other.selector

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}(selector={self.selector})"

    def __str__(self):
        return repr(self)



"""
Module with implementation of templates that enhance generator components.
"""

import random
import string
from collections.abc import Iterable
from typing import Optional

from soupsavvy.testing.generators.templates.base import BaseTemplate


class EmptyTemplate(BaseTemplate):
    """
    Template for generating an empty string.
    It's a convenient default template for always generating an empty string.

    Example
    -------
    >>> template = EmptyTemplate()
    >>> template.generate()
    ''
    """

    def generate(self) -> str:
        """
        Generates an empty string.

        Returns
        -------
        str
            An empty string.
        """
        return ""


class ConstantTemplate(BaseTemplate):
    """
    Template for generating a constant string.
    It's a convenient template for always generating the same, defined string.

    Example
    -------
    >>> template = ConstantTemplate("Hello World!")
    >>> template.generate()
    'Hello World!'
    """

    def __init__(self, value: str) -> None:
        """
        Initializes the ConstantTemplate.

        Parameters
        ----------
        value : str
            The constant string value.
        """
        self.value = str(value)

    def generate(self) -> str:
        """
        Generates the constant string.

        Returns
        -------
        str
            The constant string value defined in initializer.
        """
        return self.value


class ChoiceTemplate(BaseTemplate):
    """
    Template for randomly selecting from a list of choices.

    Example
    -------
    >>> template = ChoiceTemplate(["apple", "banana", "cherry"], seed=42)
    >>> template.generate()
    'cherry'
    """

    def __init__(self, choices: Iterable, seed: Optional[int] = None) -> None:
        """
        Initializes the ChoiceTemplate.

        Parameters
        ----------
        choices : Iterable
            Iterable of choices.
        seed : int, optional
            Random seed for reproducibility. Defaults to None, no seed is used.
        """
        self.choices = list(choices)
        self.randomizer = random.Random(seed)

    def generate(self) -> str:
        """
        Generates a random choice from the list of choices.

        Returns
        -------
        str
            The randomly selected choice.
        """
        choice = self.randomizer.choice(self.choices)
        return str(choice)


class RandomTemplate(BaseTemplate):
    """
    Template for generating a random string.

    Example
    -------
    >>> template = RandomTemplate(length=4, seed=42)
    >>> template.generate()
    'Nbrn'
    """

    _SCOPE = string.ascii_letters + string.digits

    def __init__(self, length: int = 4, seed: Optional[int] = None) -> None:
        """
        Initializes the RandomTemplate.

        Parameters
        ----------
        length : int, optional
            Length of the random string. Defaults to 4.
        seed : int, optional
            Random seed for reproducibility. Defaults to None.
        """
        self._check_length(length)
        self.length = length
        self.randomizer = random.Random(seed)

    def _check_length(self, length: int) -> None:
        """
        Checks if the input length is a valid value.

        Parameters
        ----------
        length : int
            Input length to check.

        Raises
        ------
        TypeError
            If the length is not an integer.
        ValueError
            If the length is less than 1.
        """
        if not isinstance(length, int):
            raise TypeError(f"length must be an integer, got '{type(length)}'")

        if length < 1:
            raise ValueError(f"length must be greater than 0, got '{length}'")

    def generate(self) -> str:
        """
        Generates a random string from ASCII letters and digits.

        Returns
        -------
        str
            The generated random string.
        """
        return self._generate_unique_id()

    def _generate_unique_id(self) -> str:
        """
        Generates a unique identifier string of the specified length.

        Returns
        -------
        str
            The generated unique identifier string.
        """
        new_id = "".join(
            self.randomizer.choices(
                self._SCOPE,
                k=self.length,
            )
        )
        return new_id



"""Module for custom `soupsavvy` exceptions."""


class SoupsavvyException(Exception):
    """Custom Base Exception for `soupsavvy` specific Exceptions."""


class NotTagSearcherException(SoupsavvyException, TypeError):
    """
    Exception to be raised when function excepted TagSearcher as input
    and got argument of the different, invalid type.
    """


#! SELECTORS


class SoupSelectorException(SoupsavvyException):
    """Custom Base Exception for SoupSelector specific Exceptions."""


class TagNotFoundException(SoupSelectorException):
    """
    Exception to be raised when SoupSelector element was not found in the markup.
    Raised by SoupSelector find method when strict parameter is set to True
    and target element was not found in provided bs4.Tag object.
    """


class NavigableStringException(SoupSelectorException):
    """
    Exception to be raised when NavigableString was found in find method.
    It does not return expected bs4 markup after conversion to bs4.Tag.
    When bs4.Tag find method was used with only 'string' parameter if any tag
    matches NavigableString is returned which does not have information about the tag.

    Example
    -------
    >>> nav_string = NavigableString("Hello World")
    >>> BeautifulSoup(nav_string)
    <html><body><p>Hello World</p></body></html>

    BeautifulSoup constructor always wraps NavigableString in <p> element.
    It is rather an unexpected behavior and should rather be avoided in SoupSelector
    find method, thus this exception is raised.
    """


class NotSoupSelectorException(SoupSelectorException, TypeError):
    """
    Exception to be raised when function excepted SoupSelector as input
    and got argument of the different, invalid type.
    """


class InvalidCSSSelector(SoupsavvyException):
    """
    Raised when the provided CSS selector is invalid.
    Usually raised when invalid formula was passed into 'nth' selector.

    Example
    -------
    >>> NthChild("2x + 1")
    InvalidCSSSelector
    """


#! GENERATORS


class HTMLGeneratorException(SoupsavvyException):
    """Base exception for the HTML generator module."""


class NotUniqueAttributesException(HTMLGeneratorException):
    """
    Exception raised by TagGenerator when input attributes have non-unique names.
    HTML tags cannot have multiple attributes with the same name.

    Example
    ------
    >>> TagGenerator(name="a", attrs=["href", ("href", "/endpoint")]
    NotUniqueAttributesException
    """


class VoidTagWithChildrenException(HTMLGeneratorException):
    """
    Exception raised by TagGenerator when input name of the tag is a void html tag
    like <img> or <br> and children are specified, which is not an allowed operation.

    Example
    ------
    >>> TagGenerator(name="img", children=["div", "span"])
    VoidTagWithChildrenException
    """


class EmptyNameException(HTMLGeneratorException):
    """
    Exception raised by TagGenerator when input name of the tag is empty.
    Exception raised by AttributeGenerator when input name of the attribute is empty.

    Empty string is not acceptable in these contexts.

    Example
    ------
    >>> TagGenerator(name="")
    EmptyNamException

    Example
    ------
    >>> AttributeGenerator(name="")
    EmptyNamException
    """


class InvalidTemplateException(HTMLGeneratorException):
    """
    Exception raised by AttributeGenerator when invalid template is passed as value
    or by TagGenerator when invalid template is passed as text.
    Allowed types are str, BaseTemplate and None. Passing other types is not acceptable.

    Example
    ------
    >>> AttributeGenerator(name="class", value=123)
    InvalidTemplateException

    Example
    ------
    >>> TagGenerator(name="div", text=["hello", "world"])
    InvalidTemplateException
    """


class AttributeParsingError(HTMLGeneratorException):
    """
    Exception raised by TagGenerator when data passed into attributes iterable
    contains element that couldn't be parsed into AttributeGenerator object.
    Relevant only to elements in attributes iterable that
    are not already AttributeGenerator objects.

    Example
    ------
    >>> TagGenerator(name="div", attrs=[123]) # not a string
    AttributeParsingError

    Example
    ------
    >>> TagGenerator(name="div", attrs=[("", "test")]) # empty name string
    AttributeParsingError

    Example
    ------
    >>> TagGenerator(name="div", attrs=[("class", ["hello", "soupsavvy"]]) # invalid value template
    AttributeParsingError
    """


#! OPERATIONS


class OperationException(SoupsavvyException):
    """Base exception for the BaseOperation related errors"""


class FailedOperationExecution(SoupsavvyException):
    """
    Exception raised by BaseOperation when operation execution failed and raised
    any exception.
    """


class NotOperationException(OperationException, TypeError):
    """
    Exception to be raised when function excepted BaseOperation as input
    and got argument of the different, invalid type.
    """


#! MODELS


class BaseModelException(SoupsavvyException):
    """Base exception related to models package and BaseModel class."""


class UnknownModelFieldException(BaseModelException):
    """
    Exception raised by BaseModel when model passing unknown field name
    to model constructor.
    """


class ModelNotFoundException(BaseModelException):
    """
    Exception raised by BaseModel when model scope was not found in provided tag
    and strict parameter was set to True.
    """


class FieldExtractionException(BaseModelException):
    """
    Exception raised by BaseModel when field extraction failed and related exception
    was raised by any selector step.
    """


class ScopeNotDefinedException(BaseModelException):
    """
    Exception raised by BaseModel when scope was not found in defined model class.
    __scope__ class attribute is mandatory in BaseModel derived classes.
    """


class RequiredConstraintException(BaseModelException):
    """
    Exception raised by Required field wrapper when required field was not found
    in provided tag.
    """


class FieldsNotDefinedException(BaseModelException):
    """
    Exception raised by BaseModel when fields were not found in defined model class.
    At least one field should be defined in class to be a valid model.
    """


class MissingFieldsException(BaseModelException):
    """
    Exception raised by BaseModel when parameters passed into the constructor
    do not contain all required fields defined in the model class.
    """



"""Module for utility functions for tag selectors."""

from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from itertools import chain
from typing import Optional

from bs4 import Tag


@dataclass
class TagIterator:
    """
    Wrapper class for iterating over bs4.Tag.

    Parameters
    ----------
    tag : Tag
        BeautifulSoup Tag to iterate over.
    recursive : bool, optional
        If True, iterates over all descendants, otherwise only over direct children.
        Default is True, similar to bs4 implementation.
    include_self : bool, optional
        If True, includes the tag itself in iteration, default is False.
    """

    tag: Tag
    recursive: bool = True
    include_self: bool = False

    def _get_iterator(self) -> Iterator:
        """
        Returns iterator over bs4.Tag descendants or children
        based on recursive parameter value.
        """
        return iter(self.tag.descendants if self.recursive else self.tag.children)

    def __iter__(self) -> TagIterator:
        # Resetting iterator to the beginning.
        iter_ = self._get_iterator()
        self._iter = chain([self.tag], iter_) if self.include_self else iter_
        return self

    def __next__(self) -> Tag:
        """
        Iterates over bs4.Tag skipping all strings (not bs4.Tag).
        If recursive is set to True, iterates over all descendants,
        otherwise only over direct children.
        """
        value = next(self._iter)

        while not isinstance(value, Tag):
            value = next(self._iter)

        return value


class UniqueTag:
    """
    Wrapper class for bs4.Tag to make it hashable by id.
    Hashing in bs4.Tag is implemented through hash of string representation:
    >>> str(self).__hash__()
    Two different bs4.Tag instances with the same content have the same hash,
    which is the root of the problem when operating with sets of bs4.Tag instances.

    Class uses composition over inheritance which makes it responsible
    only for hashing and equality check.

    Class is used in Selectors that use operations on sets of bs4.Tag instances.
    """

    def __init__(self, tag: Tag):
        """
        Initializes UniqueTag instance.

        Parameters
        ----------
        tag : bs4.Tag
            BeautifulSoup Tag instance to wrap.
        """
        self.tag = tag

    def __hash__(self):
        """Hashes bs4.Tag instance by id."""
        return id(self.tag)

    def __eq__(self, other):
        """
        Checks equality of itself with another object.

        For object to be equal to UniqueTag, it should be an instance of UniqueTag
        and have the same hash value, meaning it has to wrap the same bs4.Tag instance.
        In any other case, returns False.
        """
        return isinstance(other, UniqueTag) and hash(self) == hash(other)

    def __str__(self):
        """Returns string representation of UniqueTag instance."""
        return f"{self.__class__.__name__}({id(self.tag)})"


class TagResultSet:
    """
    TagResultSet class is collection that stores and manages results of find_all
    method of selectors. Prerequisites for returned results are:
    * bs4.Tag instances are unique
    * the order of results == order of their appearance in html

    This components consumes list of bs4.Tag instances and provides methods
    for fetching unique results with preserved order.
    It provides operations on sets of results like intersection and union.
    """

    # constants used inside the class
    _ORDER_ATTR = "_order"
    _IS_BASE = "_base"

    def __init__(self, tags: Optional[list[Tag]] = None) -> None:
        """
        Initializes TagResultSet instance.

        Parameters
        ----------
        tags : list[Tag], optional
            List of bs4.Tag instances to store in the collection.
            Default is None, which initializes empty collection.
        """
        self._tags = tags or []

    def fetch(self, n: Optional[int] = None) -> list[Tag]:
        """
        Fetches n first unique bs4.Tag instances from collection.
        Ensures that the order of the initial list is preserved.

        Parameters
        ----------
        n : int, optional
            Number of bs4.Tag instances to fetch.
            If default None, fetches all unique bs4.Tag instances.

        Returns
        -------
        list[Tag]
            List of bs4.Tag instances fetched from collection.
        """
        set_ = self._to_set(base=True)
        ordered = self._sort(set_)
        return ordered[:n]

    def _to_set(self, base: bool) -> set[UniqueTag]:
        """
        Converts list of bs4.Tags from collection to set of UniqueTag instances.

        Parameters
        ----------
        base : bool
            If True, sets the tag as base, otherwise as non-base.
            If TagResultSet is used in set operations as a base, it should be True.

        Returns
        -------
        set[UniqueTag]
            Set of UniqueTag instances with set helper attributes.
        """
        # converting to UniqueTag instances due to known hashing issue
        tags = [UniqueTag(tag) for tag in self._tags]

        for i, tag in enumerate(tags):
            # setting attributes used for restoring order
            setattr(tag, self._ORDER_ATTR, i)
            setattr(tag, self._IS_BASE, int(base))

        return set(tags)

    def _sort(self, it: Iterable[UniqueTag]) -> list[Tag]:
        """
        Sorts an iterable of UniqueTag instances by order and base attributes.

        Parameters
        ----------
        it : Iterable[UniqueTag]
            Iterable of UniqueTag instances to sort.

        Returns
        -------
        list[Tag]
            List of bs4.Tag instances sorted by order and base attributes.
        """
        return [
            unique.tag
            for unique in sorted(
                it,
                key=lambda x: (
                    # Sorting by base descending - base goes first
                    -getattr(x, self.__class__._IS_BASE),
                    # Sorting by order ascending
                    getattr(x, self._ORDER_ATTR),
                ),
            )
        ]

    def __and__(self, other: TagResultSet) -> TagResultSet:
        """
        Performs an intersection operation on two TagResultSet instances
        with current instance as a base,
        preserving the order of tags from the base instance.

        Parameters
        ----------
        other : TagResultSet
            TagResultSet instance to perform intersection with.

        Example
        -------
        >>> base = TagResultSet([x, y, b])
        >>> other = TagResultSet([c, y, x])
        >>> base & other
        TagResultSet([x, y])

        Returns
        -------
        TagResultSet
            New TagResultSet instance with results of intersection operation.
        """
        base = self._to_set(base=True)
        right = other._to_set(base=False)
        # if set intersection, objects are taken from right operant, which messes up the order
        intersection = [obj for obj in base if obj in right]
        ordered = self._sort(intersection)
        return TagResultSet(ordered)

    def __or__(self, other: TagResultSet) -> TagResultSet:
        """
        Performs a union operation on two TagResultSet instances
        with current instance list of tags as a base, appending new tags
        from other instance at the end of the list.

        Parameters
        ----------
        other : TagResultSet
            TagResultSet instance to perform union with.

        Example
        -------
        >>> base = TagResultSet([x, y, b])
        >>> other = TagResultSet([c, y, x])
        >>> base | other
        TagResultSet([x, y, b, c])

        Returns
        -------
        TagResultSet
            New TagResultSet instance with results of union operation.
        """
        base = self._to_set(base=True)
        right = other._to_set(base=False)
        updated = base | right
        ordered = self._sort(updated)
        return TagResultSet(ordered)

    def __sub__(self, other: TagResultSet) -> TagResultSet:
        """
        Performs a difference operation on two TagResultSet instances
        with current instance as a base,
        preserving the order of tags from the base instance.

        Parameters
        ----------
        other : TagResultSet
            TagResultSet instance to perform difference with.

        Example
        -------
        >>> base = TagResultSet([x, y, b])
        >>> other = TagResultSet([c, y, x])
        >>> base - other
        TagResultSet([b])

        Returns
        -------
        TagResultSet
            New TagResultSet instance with results of difference operation.
        """
        base = self._to_set(base=True)
        right = other._to_set(base=False)
        difference = base - right
        ordered = self._sort(difference)
        return TagResultSet(ordered)

    def symmetric_difference(self, other: TagResultSet) -> TagResultSet:
        """
        Performs a symmetric difference operation on two TagResultSet instances
        with current instance as a base,
        preserving the order of tags from the base instance.

        Parameters
        ----------
        other : TagResultSet
            TagResultSet instance to perform symmetric difference with.

        Example
        -------
        >>> base = TagResultSet([x, y, b])
        >>> other = TagResultSet([c, y, x])
        >>> base.symmetric_difference(other)
        TagResultSet([b, c])

        Returns
        -------
        TagResultSet
            New TagResultSet instance with results of symmetric difference operation.
        """
        base = self._to_set(base=True)
        right = other._to_set(base=False)
        symmetric_diff = base.symmetric_difference(right)
        ordered = self._sort(symmetric_diff)
        return TagResultSet(ordered)

    def __len__(self) -> int:
        """Returns the number of bs4.Tag instances in the collection."""
        return len(self._tags)

    def __bool__(self) -> bool:
        """Returns True if collection is not empty, otherwise False."""
        return len(self) > 0
